@@ -7,7 +7,7 @@
 
    Author: Ilan Schnell
 */
-#define BITARRAY_VERSION  "1.4.0"
+#define BITARRAY_VERSION  "1.4.2"
 
 #define PY_SSIZE_T_CLEAN
 #include "Python.h"
@@ -18,6 +18,9 @@
 
 #ifdef IS_PY3K
 #define Py_TPFLAGS_HAVE_WEAKREFS  0
+#else
+/* this macro was introduced in Python 3.3 */
+#define Py_MIN(x, y)  (((x) > (y)) ? (y) : (x))
 #endif
 
 #if PY_MAJOR_VERSION == 3 || (PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION == 7)
@@ -71,6 +74,10 @@ static int default_endian = ENDIAN_BIG;
 #define BITMASK(endian, i)  \
     (((char) 1) << ((endian) == ENDIAN_LITTLE ? ((i) % 8) : (7 - (i) % 8)))
 
+/* This (bytes) block size is used when reading/writing blocks of bytes
+   from files. */
+#define BLOCKSIZE  65536
+
 /* ------------ low level access to bits in bitarrayobject ------------- */
 
 #ifndef NDEBUG
@@ -131,7 +138,7 @@ resize(bitarrayobject *self, idx_t nbits)
     newsize = (Py_ssize_t) BYTES(nbits);
 
     if (newsize == size) {
-        /* the memory size hasn't changed - bypass almost everything */
+        /* the memory size hasn't changed - bypass everything */
         self->nbits = nbits;
         return 0;
     }
@@ -167,12 +174,10 @@ resize(bitarrayobject *self, idx_t nbits)
         /* When resizing an empty bitarray, we want at least 4 bytes. */
         new_allocated = 4;
 
-    /* Over-allocate unless (one of the following):
-         - The (previous) size is zero, as we often
-           extend an empty array on creation.
-         - The size increase is very large.
-         - The size is decreasing. */
-    else if (!(size == 0 || newsize > size + 65536 || newsize < size))
+    /* Over-allocate when the (previous) size is non-zero (as we often
+       extend an empty array on creation) and the size is actually
+       increasing. */
+    else if (size != 0 && newsize > size)
         /* This over-allocates proportional to the bitarray size, making
            room for additional growth.
            The growth pattern is:  0, 4, 8, 16, 25, 34, 44, 54, 65, 77, ...
@@ -260,7 +265,7 @@ copy_n(bitarrayobject *self, idx_t a,
        self to self. */
     if (self->endian == other->endian && a % 8 == 0 && b % 8 == 0 && n >= 8)
     {
-        const Py_ssize_t bytes = (Py_ssize_t) n / 8;
+        const size_t bytes = n / 8;
         const idx_t bits = BITS(bytes);
 
         assert(bits <= n && n < bits + 8);
@@ -317,37 +322,45 @@ insert_n(bitarrayobject *self, idx_t start, idx_t n)
     return 0;
 }
 
-/* sets ususet bits to 0, i.e. the ones in the last byte (if any),
-   and return the number of bits set -- self->nbits is unchanged */
+/* sets unused padding bits (within last byte of buffer) to 0,
+   and return the number of padding bits -- self->nbits is unchanged */
 static int
 setunused(bitarrayobject *self)
 {
-    const idx_t n = BITS(Py_SIZE(self));
-    idx_t i;
+    idx_t n, i;
 
+    if (self->nbits % 8 == 0)
+        return 0;
+
+    n = BITS(Py_SIZE(self));    /* number of bits in buffer */
     for (i = self->nbits; i < n; i++)
         setbit(self, i, 0);
-    assert(n - self->nbits < 8);
-    return n - self->nbits;
+    assert(0 < n - self->nbits && n - self->nbits < 8);
+    return (int) (n - self->nbits);
 }
 
-/* repeat self n times */
+static void
+invert(bitarrayobject *self)
+{
+    Py_ssize_t nbytes = Py_SIZE(self), i;
+
+    for (i = 0; i < nbytes; i++)
+        self->ob_item[i] = ~self->ob_item[i];
+}
+
+/* repeat self n times (negative n is treated as 0) */
 static int
 repeat(bitarrayobject *self, idx_t n)
 {
     idx_t nbits, i;
 
-    if (n <= 0) {
-        if (resize(self, 0) < 0)
-            return -1;
-    }
-    if (n > 1) {
-        nbits = self->nbits;
-        if (resize(self, nbits * n) < 0)
-            return -1;
-        for (i = 1; i < n; i++)
-            copy_n(self, i * nbits, self, 0, nbits);
-    }
+    if (n <= 0)
+        n = 0;
+    nbits = self->nbits;
+    if (resize(self, n * nbits) < 0)
+        return -1;
+    for (i = 1; i < n; i++)
+        copy_n(self, i * nbits, self, 0, nbits);
     return 0;
 }
 
@@ -391,7 +404,7 @@ bitwise(bitarrayobject *self, PyObject *arg, enum op_type oper)
         for (i = 0; i < n; i++)
             self->ob_item[i] ^= other->ob_item[i];
         break;
-    default:  /* should never happen */
+    default:  /* cannot happen */
         return -1;
     }
     return 0;
@@ -415,7 +428,7 @@ setrange(bitarrayobject *self, idx_t start, idx_t stop, int val)
         for (i = start; i < BITS(byte_start); i++)
             setbit(self, i, val);
         memset(self->ob_item + byte_start, val ? 0xff : 0x00,
-               byte_stop - byte_start);
+               (size_t) (byte_stop - byte_start));
         for (i = BITS(byte_stop); i < stop; i++)
             setbit(self, i, val);
     }
@@ -425,42 +438,6 @@ setrange(bitarrayobject *self, idx_t start, idx_t stop, int val)
     }
 }
 
-static void
-invert(bitarrayobject *self)
-{
-    Py_ssize_t i;
-
-    for (i = 0; i < Py_SIZE(self); i++)
-        self->ob_item[i] = ~self->ob_item[i];
-}
-
-/* reverse the order of bits in each byte of the buffer */
-static void
-bytereverse(bitarrayobject *self)
-{
-    static char trans[256];
-    static int setup = 0;
-    Py_ssize_t i;
-
-    if (!setup) {
-        /* setup translation table, which maps each byte to it's reversed:
-           trans = {0, 128, 64, 192, 32, 160, ..., 255} */
-        int j, k;
-
-        for (k = 0; k < 256; k++) {
-            trans[k] = 0x00;
-            for (j = 0; j < 8; j++)
-                if (1 << (7 - j) & k)
-                    trans[k] |= 1 << j;
-        }
-        setup = 1;
-    }
-
-    setunused(self);
-    for (i = 0; i < Py_SIZE(self); i++)
-        self->ob_item[i] = trans[(unsigned char) self->ob_item[i]];
-}
-
 static unsigned char bitcount_lookup[256] = {
     0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
     1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
@@ -585,30 +562,6 @@ append_item(bitarrayobject *self, PyObject *item)
     return set_item(self, self->nbits - 1, item);
 }
 
-static PyObject *
-unpack(bitarrayobject *self, char zero, char one)
-{
-    PyObject *result;
-    Py_ssize_t i;
-    char *str;
-
-    if (self->nbits > PY_SSIZE_T_MAX) {
-        PyErr_SetString(PyExc_OverflowError, "bitarray too large to unpack");
-        return NULL;
-    }
-    str = (char *) PyMem_Malloc((size_t) self->nbits);
-    if (str == NULL) {
-        PyErr_NoMemory();
-        return NULL;
-    }
-    for (i = 0; i < self->nbits; i++) {
-        str[i] = GETBIT(self, i) ? one : zero;
-    }
-    result = PyBytes_FromStringAndSize(str, (Py_ssize_t) self->nbits);
-    PyMem_Free((void *) str);
-    return result;
-}
-
 static int
 extend_bitarray(bitarrayobject *self, bitarrayobject *other)
 {
@@ -694,15 +647,8 @@ extend_tuple(bitarrayobject *self, PyObject *tuple)
     return 0;
 }
 
-/* extend_bytes(): extend the bitarray from a PyBytes object, where each
-   whole character is converted to a single bit */
-enum conv_t {
-    BYTES_01,    /*  '0' -> 0    '1'  -> 1   no other characters allowed */
-    BYTES_RAW,   /*  0x00 -> 0   other -> 1                              */
-};
-
 static int
-extend_bytes(bitarrayobject *self, PyObject *bytes, enum conv_t conv)
+extend_01(bitarrayobject *self, PyObject *bytes)
 {
     Py_ssize_t nbytes, i;
     char c, *data;
@@ -717,25 +663,14 @@ extend_bytes(bitarrayobject *self, PyObject *bytes, enum conv_t conv)
         return -1;
 
     data = PyBytes_AsString(bytes);
-
     for (i = 0; i < nbytes; i++) {
         c = data[i];
-        /* depending on conv, map c to bit */
-        switch (conv) {
-        case BYTES_01:
-            switch (c) {
-            case '0': vi = 0; break;
-            case '1': vi = 1; break;
-            default:
-                PyErr_Format(PyExc_ValueError,
-                             "character must be '0' or '1', found '%c'", c);
-                return -1;
-            }
-            break;
-        case BYTES_RAW:
-            vi = c ? 1 : 0;
-            break;
-        default:  /* should never happen */
+        switch (c) {
+        case '0': vi = 0; break;
+        case '1': vi = 1; break;
+        default:
+            PyErr_Format(PyExc_ValueError,
+                         "character must be '0' or '1', found '%c'", c);
             return -1;
         }
         setbit(self, self->nbits - nbytes + i, vi);
@@ -743,25 +678,6 @@ extend_bytes(bitarrayobject *self, PyObject *bytes, enum conv_t conv)
     return 0;
 }
 
-static int
-extend_rawbytes(bitarrayobject *self, PyObject *bytes)
-{
-    Py_ssize_t nbytes;
-    char *data;
-
-    assert(PyBytes_Check(bytes) && self->nbits % 8 == 0);
-    nbytes = PyBytes_Size(bytes);
-    if (nbytes == 0)
-        return 0;
-
-    if (resize(self, self->nbits + BITS(nbytes)) < 0)
-        return -1;
-
-    data = PyBytes_AsString(bytes);
-    memcpy(self->ob_item + (Py_SIZE(self) - nbytes), data, nbytes);
-    return 0;
-}
-
 static int
 extend_dispatch(bitarrayobject *self, PyObject *obj)
 {
@@ -778,14 +694,27 @@ extend_dispatch(bitarrayobject *self, PyObject *obj)
     if (PyTuple_Check(obj))                                  /* tuple */
         return extend_tuple(self, obj);
 
-    if (PyBytes_Check(obj))                               /* bytes 01 */
-        return extend_bytes(self, obj, BYTES_01);
+    if (PyBytes_Check(obj)) {                             /* bytes 01 */
+        /* This case is used on Python 2.  However, it should have never
+           been here for Python 3, as it allows bitarray(b'01101011') */
+#ifdef IS_PY3K
+        if (PyErr_WarnEx(PyExc_DeprecationWarning,
+                         "cannot extend from 'bytes', "
+                         "use .pack() or .frombytes() instead", 1) < 0)
+            return -1;
+#endif
+        return extend_01(self, obj);
+    }
 
     if (PyUnicode_Check(obj)) {                /* (unicode) string 01 */
         PyObject *bytes;
+
         bytes = PyUnicode_AsEncodedString(obj, NULL, NULL);
-        ret = extend_bytes(self, bytes, BYTES_01);
-        Py_DECREF(bytes);
+        if (bytes == NULL)
+            return -1;
+        assert(PyBytes_Check(bytes));
+        ret = extend_01(self, bytes);
+        Py_DECREF(bytes);  /* drop bytes */
         return ret;
     }
 
@@ -804,6 +733,31 @@ extend_dispatch(bitarrayobject *self, PyObject *obj)
     return -1;
 }
 
+static PyObject *
+unpack(bitarrayobject *self, char zero, char one, const char *fmt)
+{
+    PyObject *result;
+    Py_ssize_t i;
+    char *str;
+
+    if (self->nbits > PY_SSIZE_T_MAX) {
+        PyErr_SetString(PyExc_OverflowError, "bitarray too large to unpack");
+        return NULL;
+    }
+
+    str = (char *) PyMem_Malloc((size_t) self->nbits);
+    if (str == NULL) {
+        PyErr_NoMemory();
+        return NULL;
+    }
+    for (i = 0; i < self->nbits; i++)
+        str[i] = GETBIT(self, i) ? one : zero;
+
+    result = Py_BuildValue(fmt, str, (Py_ssize_t) self->nbits);
+    PyMem_Free((void *) str);
+    return result;
+}
+
 /* --------- helper functions not involving bitarrayobjects ------------ */
 
 #ifdef IS_PY3K
@@ -951,7 +905,7 @@ slice_GetIndicesEx(PySliceObject *r, idx_t length,
 }
 
 /**************************************************************************
-                         Implementation of API methods
+                     Implementation of bitarray methods
  **************************************************************************/
 
 static PyObject *
@@ -986,7 +940,8 @@ bitarray_copy(bitarrayobject *self)
     if (res == NULL)
         return NULL;
 
-    memcpy(((bitarrayobject *) res)->ob_item, self->ob_item, Py_SIZE(self));
+    memcpy(((bitarrayobject *) res)->ob_item, self->ob_item,
+           (size_t) Py_SIZE(self));
     return res;
 }
 
@@ -1063,11 +1018,11 @@ bitarray_extend(bitarrayobject *self, PyObject *obj)
 }
 
 PyDoc_STRVAR(extend_doc,
-"extend(iterable, /)\n\
+"extend(iterable or string, /)\n\
 \n\
-Append bits to the end of the bitarray.  The objects which can be passed\n\
-to this method are the same iterable objects which can given to a bitarray\n\
-object upon initialization.");
+Extend bitarray by appending the truth value of each element given\n\
+by iterable.  If a string is provided, each `0` and `1` are appended\n\
+as bits.");
 
 
 static PyObject *
@@ -1203,7 +1158,7 @@ bitarray_append(bitarrayobject *self, PyObject *v)
 PyDoc_STRVAR(append_doc,
 "append(item, /)\n\
 \n\
-Append the value `bool(item)` to the end of the bitarray.");
+Append the truth value `bool(item)` to the end of the bitarray.");
 
 
 static PyObject *
@@ -1251,7 +1206,7 @@ bitarray_reduce(bitarrayobject *self)
         goto error;
     }
     data[0] = (char) setunused(self);
-    memcpy(data + 1, self->ob_item, nbytes);
+    memcpy(data + 1, self->ob_item, (size_t) nbytes);
     repr = PyBytes_FromStringAndSize(data, nbytes + 1);
     if (repr == NULL)
         goto error;
@@ -1271,29 +1226,29 @@ static PyObject *
 bitarray_reverse(bitarrayobject *self)
 {
     const idx_t m = self->nbits - 1;   /* index of max item of self */
-    bitarrayobject *t;    /* temp bitarray to store lower half of self */
+    PyObject *t;       /* temp bitarray to store lower half of self */
     idx_t i;
 
     if (self->nbits < 2)        /* nothing needs to be done */
         Py_RETURN_NONE;
 
-    t = (bitarrayobject *) newbitarrayobject(Py_TYPE(self),
-                                             self->nbits / 2, self->endian);
+    t = newbitarrayobject(Py_TYPE(self), self->nbits / 2, self->endian);
     if (t == NULL)
         return NULL;
 
+#define tt  ((bitarrayobject *) t)
     /* copy lower half of array into temporary array */
-    memcpy(t->ob_item, self->ob_item, Py_SIZE(t));
+    memcpy(tt->ob_item, self->ob_item, (size_t) Py_SIZE(tt));
 
-    /* reverse the upper half onto the lower half. */
-    for (i = 0; i < t->nbits; i++)
+    /* reverse upper half onto the lower half. */
+    for (i = 0; i < tt->nbits; i++)
         setbit(self, i, GETBIT(self, m - i));
 
-    /* revert the stored away lower half onto the upper half. */
-    for (i = 0; i < t->nbits; i++)
-        setbit(self, m - i, GETBIT(t, i));
-
-    bitarray_dealloc(t);
+    /* reverse the stored away lower half onto the upper half of self. */
+    for (i = 0; i < tt->nbits; i++)
+        setbit(self, m - i, GETBIT(tt, i));
+#undef tt
+    Py_DECREF(t);
     Py_RETURN_NONE;
 }
 
@@ -1337,7 +1292,28 @@ i.e. convert each 1-bit into a 0-bit and vice versa.");
 static PyObject *
 bitarray_bytereverse(bitarrayobject *self)
 {
-    bytereverse(self);
+    static char trans[256];
+    static int setup = 0;
+    Py_ssize_t i;
+
+    if (!setup) {
+        /* setup translation table, which maps each byte to it's reversed:
+           trans = {0, 128, 64, 192, 32, 160, ..., 255} */
+        int j, k;
+
+        for (k = 0; k < 256; k++) {
+            trans[k] = 0x00;
+            for (j = 0; j < 8; j++)
+                if (1 << (7 - j) & k)
+                    trans[k] |= 1 << j;
+        }
+        setup = 1;
+    }
+
+    setunused(self);
+    for (i = 0; i < Py_SIZE(self); i++)
+        self->ob_item[i] = trans[(unsigned char) self->ob_item[i]];
+
     Py_RETURN_NONE;
 }
 
@@ -1358,7 +1334,7 @@ bitarray_setall(bitarrayobject *self, PyObject *v)
     if (vi < 0)
         return NULL;
 
-    memset(self->ob_item, vi ? 0xff : 0x00, Py_SIZE(self));
+    memset(self->ob_item, vi ? 0xff : 0x00, (size_t) Py_SIZE(self));
     Py_RETURN_NONE;
 }
 
@@ -1399,223 +1375,6 @@ PyDoc_STRVAR(sort_doc,
 Sort the bits in the array (in-place).");
 
 
-/* since too many details differ between the Python 2 and 3 implementation
-   of this function, we choose to have two separate function implementation,
-   even though this means some of the code is duplicated in the two versions
-*/
-#ifdef IS_PY3K
-static PyObject *
-bitarray_fromfile(bitarrayobject *self, PyObject *args)
-{
-    PyObject *f;
-    Py_ssize_t newsize, nbytes = -1;
-    PyObject *reader, *rargs, *result;
-    size_t nread;
-    idx_t t, p;
-
-    if (!PyArg_ParseTuple(args, "O|n:fromfile", &f, &nbytes))
-        return NULL;
-
-    if (nbytes == 0)
-        Py_RETURN_NONE;
-
-    reader = PyObject_GetAttrString(f, "read");
-    if (reader == NULL)
-    {
-        PyErr_SetString(PyExc_TypeError,
-                        "first argument must be an open file");
-        return NULL;
-    }
-    rargs = Py_BuildValue("(n)", nbytes);
-    if (rargs == NULL) {
-        Py_DECREF(reader);
-        return NULL;
-    }
-    result = PyEval_CallObject(reader, rargs);
-    if (result != NULL) {
-        if (!PyBytes_Check(result)) {
-            PyErr_SetString(PyExc_TypeError,
-                            "first argument must be an open file");
-            Py_DECREF(result);
-            Py_DECREF(rargs);
-            Py_DECREF(reader);
-            return NULL;
-        }
-        nread = PyBytes_Size(result);
-
-        t = self->nbits;
-        p = setunused(self);
-        self->nbits += p;
-
-        newsize = Py_SIZE(self) + nread;
-
-        if (resize(self, BITS(newsize)) < 0) {
-            Py_DECREF(result);
-            Py_DECREF(rargs);
-            Py_DECREF(reader);
-            return NULL;
-        }
-        memcpy(self->ob_item + (Py_SIZE(self) - nread),
-               PyBytes_AS_STRING(result), nread);
-
-        if (nbytes > 0 && nread < (size_t) nbytes) {
-            PyErr_SetString(PyExc_EOFError, "not enough items read");
-            return NULL;
-        }
-        if (delete_n(self, t, p) < 0)
-            return NULL;
-        Py_DECREF(result);
-    }
-
-    Py_DECREF(rargs);
-    Py_DECREF(reader);
-    Py_RETURN_NONE;
-}
-#else  /* Python 2 */
-static PyObject *
-bitarray_fromfile(bitarrayobject *self, PyObject *args)
-{
-    PyObject *f;
-    FILE *fp;
-    Py_ssize_t newsize, nbytes = -1;
-    size_t nread;
-    idx_t t, p;
-    long cur;
-
-    if (!PyArg_ParseTuple(args, "O|n:fromfile", &f, &nbytes))
-        return NULL;
-
-    fp = PyFile_AsFile(f);
-    if (fp == NULL) {
-        PyErr_SetString(PyExc_TypeError,
-                        "first argument must be an open file");
-        return NULL;
-    }
-
-    /* find number of bytes till EOF */
-    if (nbytes < 0) {
-        if ((cur = ftell(fp)) < 0)
-            goto EOFerror;
-
-        if (fseek(fp, 0L, SEEK_END) || (nbytes = ftell(fp)) < 0)
-            goto EOFerror;
-
-        nbytes -= cur;
-        if (fseek(fp, cur, SEEK_SET)) {
-        EOFerror:
-            PyErr_SetString(PyExc_EOFError, "could not find EOF");
-            return NULL;
-        }
-    }
-    if (nbytes == 0)
-        Py_RETURN_NONE;
-
-    /* file exists and there are more than zero bytes to read */
-    t = self->nbits;
-    p = setunused(self);
-    self->nbits += p;
-
-    newsize = Py_SIZE(self) + nbytes;
-    if (resize(self, BITS(newsize)) < 0)
-        return NULL;
-
-    nread = fread(self->ob_item + (Py_SIZE(self) - nbytes), 1, nbytes, fp);
-    if (nread < (size_t) nbytes) {
-        newsize -= nbytes - nread;
-        if (resize(self, BITS(newsize)) < 0)
-            return NULL;
-        PyErr_SetString(PyExc_EOFError, "not enough items in file");
-        return NULL;
-    }
-
-    if (delete_n(self, t, p) < 0)
-        return NULL;
-    Py_RETURN_NONE;
-}
-#endif
-
-PyDoc_STRVAR(fromfile_doc,
-"fromfile(f, n=<till EOF>, /)\n\
-\n\
-Read n bytes from the file object f and append them to the bitarray\n\
-interpreted as machine values.  When n is omitted, as many bytes are\n\
-read until EOF is reached.");
-
-
-/* since too many details differ between the Python 2 and 3 implementation
-   of this function, we choose to have two separate function implementation
-*/
-#ifdef IS_PY3K
-static PyObject *
-bitarray_tofile(bitarrayobject *self, PyObject *f)
-{
-    PyObject *writer, *value, *args, *result;
-
-    if (f == NULL) {
-        PyErr_SetString(PyExc_TypeError, "writeobject with NULL file");
-        return NULL;
-    }
-    writer = PyObject_GetAttrString(f, "write");
-    if (writer == NULL)
-        return NULL;
-    setunused(self);
-    value = PyBytes_FromStringAndSize(self->ob_item, Py_SIZE(self));
-    if (value == NULL) {
-        Py_DECREF(writer);
-        return NULL;
-    }
-    args = PyTuple_Pack(1, value);
-    if (args == NULL) {
-        Py_DECREF(value);
-        Py_DECREF(writer);
-        return NULL;
-    }
-    result = PyEval_CallObject(writer, args);
-    Py_DECREF(args);
-    Py_DECREF(value);
-    Py_DECREF(writer);
-    if (result == NULL)
-    {
-        PyErr_SetString(PyExc_TypeError, "open file expected");
-        return NULL;
-    }
-    Py_DECREF(result);
-    Py_RETURN_NONE;
-}
-#else  /* Python 2 */
-static PyObject *
-bitarray_tofile(bitarrayobject *self, PyObject *f)
-{
-    FILE *fp;
-
-    fp = PyFile_AsFile(f);
-    if (fp == NULL) {
-        PyErr_SetString(PyExc_TypeError, "open file expected");
-        return NULL;
-    }
-    if (Py_SIZE(self) == 0)
-        Py_RETURN_NONE;
-
-    setunused(self);
-    if (fwrite(self->ob_item, 1, Py_SIZE(self), fp) !=
-        (size_t) Py_SIZE(self))
-    {
-        PyErr_SetFromErrno(PyExc_IOError);
-        clearerr(fp);
-        return NULL;
-    }
-    Py_RETURN_NONE;
-}
-#endif
-
-PyDoc_STRVAR(tofile_doc,
-"tofile(f, /)\n\
-\n\
-Write all bits (as machine values) to the file object f.\n\
-When the length of the bitarray is not a multiple of 8,\n\
-the remaining bits (1..7) are set to 0.");
-
-
 static PyObject *
 bitarray_tolist(bitarrayobject *self)
 {
@@ -1636,35 +1395,42 @@ bitarray_tolist(bitarrayobject *self)
 PyDoc_STRVAR(tolist_doc,
 "tolist() -> list\n\
 \n\
-Return an ordinary list with the items in the bitarray.\n\
+Return a list with the items (False or True) in the bitarray.\n\
 Note that the list object being created will require 32 or 64 times more\n\
-memory than the bitarray object, which may cause a memory error if the\n\
-bitarray is very large.\n\
-Also note that to extend a bitarray with elements from a list,\n\
-use the extend method.");
+memory (depending on the machine architecture) than the bitarray object,\n\
+which may cause a memory error if the bitarray is very large.");
 
 
 static PyObject *
 bitarray_frombytes(bitarrayobject *self, PyObject *bytes)
 {
+    Py_ssize_t nbytes;
     idx_t t, p;
 
     if (!PyBytes_Check(bytes)) {
         PyErr_SetString(PyExc_TypeError, "bytes expected");
         return NULL;
     }
+    nbytes = PyBytes_GET_SIZE(bytes);
+    if (nbytes == 0)
+        Py_RETURN_NONE;
 
     /* Before we extend the raw bytes with the new data, we need to store
        the current size and pad the last byte, as our bitarray size might
        not be a multiple of 8.  After extending, we remove the padding
-       bits again.  The same is done in bitarray_fromfile().
+       bits again.
     */
     t = self->nbits;
     p = setunused(self);
     self->nbits += p;
+    assert(self->nbits % 8 == 0);
 
-    if (extend_rawbytes(self, bytes) < 0)
+    if (resize(self, self->nbits + BITS(nbytes)) < 0)
         return NULL;
+
+    memcpy(self->ob_item + (Py_SIZE(self) - nbytes),
+           PyBytes_AsString(bytes), (size_t) nbytes);
+
     if (delete_n(self, t, p) < 0)
         return NULL;
     Py_RETURN_NONE;
@@ -1673,7 +1439,8 @@ bitarray_frombytes(bitarrayobject *self, PyObject *bytes)
 PyDoc_STRVAR(frombytes_doc,
 "frombytes(bytes, /)\n\
 \n\
-Append from a byte string, interpreted as machine values.");
+Extend bitarray with raw bytes.  That is, each append byte will add eight\n\
+bits to the bitarray.");
 
 
 static PyObject *
@@ -1691,31 +1458,99 @@ When the length of the bitarray is not a multiple of 8, the few remaining\n\
 bits (1..7) are considered to be 0.");
 
 
+static PyObject *
+bitarray_fromfile(bitarrayobject *self, PyObject *args)
+{
+    PyObject *bytes, *f, *res;
+    Py_ssize_t nblock, nread = 0, nbytes = -1;
+    int not_enough_bytes;
+
+    if (!PyArg_ParseTuple(args, "O|n:fromfile", &f, &nbytes))
+        return NULL;
+
+    if (nbytes < 0)  /* read till EOF */
+        nbytes = PY_SSIZE_T_MAX;
+
+    while (nread < nbytes) {
+        nblock = Py_MIN(nbytes - nread, BLOCKSIZE);
+        bytes = PyObject_CallMethod(f, "read", "n", nblock);
+        if (bytes == NULL)
+            return NULL;
+        if (!PyBytes_Check(bytes)) {
+            Py_DECREF(bytes);
+            PyErr_SetString(PyExc_TypeError, "read() didn't return bytes");
+            return NULL;
+        }
+        not_enough_bytes = (PyBytes_GET_SIZE(bytes) < nblock);
+        nread += PyBytes_GET_SIZE(bytes);
+        assert(nread >= 0 && nread <= nbytes);
+
+        res = bitarray_frombytes(self, bytes);
+        Py_DECREF(bytes);
+        if (res == NULL)
+            return NULL;
+        Py_DECREF(res);  /* drop frombytes result */
+
+        if (not_enough_bytes) {
+            if (nbytes == PY_SSIZE_T_MAX)  /* read till EOF */
+                break;
+            PyErr_SetString(PyExc_EOFError, "not enough bytes to read");
+            return NULL;
+        }
+    }
+    Py_RETURN_NONE;
+}
+
+PyDoc_STRVAR(fromfile_doc,
+"fromfile(f, n=-1, /)\n\
+\n\
+Extend bitarray with up to n bytes read from the file object f.\n\
+When n is omitted or negative, reads all data until EOF.\n\
+When n is provided and positions but exceeds the data available,\n\
+EOFError is raised (but the available data is still read and appended.");
+
+
+static PyObject *
+bitarray_tofile(bitarrayobject *self, PyObject *f)
+{
+    Py_ssize_t size, nbytes = Py_SIZE(self);
+    Py_ssize_t offset;
+    PyObject *res;
+
+    setunused(self);
+    for (offset = 0; offset < nbytes; offset += BLOCKSIZE) {
+        size = Py_MIN(nbytes - offset, BLOCKSIZE);
+        assert(size >=0 && offset + size <= nbytes);
+        /* basically: f.write(memoryview(self)[offset:offset+size] */
+        res = PyObject_CallMethod(f, "write",
+                                  PY_MAJOR_VERSION == 2 ? "s#" : "y#",
+                                  self->ob_item + offset, size);
+        if (res == NULL)
+            return NULL;
+        Py_DECREF(res);  /* drop write result */
+    }
+    Py_RETURN_NONE;
+}
+
+PyDoc_STRVAR(tofile_doc,
+"tofile(f, /)\n\
+\n\
+Write the byte representation of the bitarray to the file object f.\n\
+When the length of the bitarray is not a multiple of 8,\n\
+the remaining bits (1..7) are set to 0.");
+
+
 static PyObject *
 bitarray_to01(bitarrayobject *self)
 {
-#ifdef IS_PY3K
-    PyObject *string;
-    PyObject *unpacked;
-
-    unpacked = unpack(self, '0', '1');
-    if (unpacked == NULL)
-        return NULL;
-    string = PyUnicode_FromEncodedObject(unpacked, NULL, NULL);
-    Py_DECREF(unpacked);
-    return string;
-#else
-    return unpack(self, '0', '1');
-#endif
+    return unpack(self, '0', '1', "s#");
 }
 
 PyDoc_STRVAR(to01_doc,
 "to01() -> str\n\
 \n\
 Return a string containing '0's and '1's, representing the bits in the\n\
-bitarray object.\n\
-Note: To extend a bitarray from a string containing '0's and '1's,\n\
-use the extend method.");
+bitarray object.");
 
 
 static PyObject *
@@ -1728,7 +1563,7 @@ bitarray_unpack(bitarrayobject *self, PyObject *args, PyObject *kwds)
                                      &zero, &one))
         return NULL;
 
-    return unpack(self, zero, one);
+    return unpack(self, zero, one, PY_MAJOR_VERSION == 3 ? "y#" : "s#");
 }
 
 PyDoc_STRVAR(unpack_doc,
@@ -1741,13 +1576,24 @@ using the specified mapping.");
 static PyObject *
 bitarray_pack(bitarrayobject *self, PyObject *bytes)
 {
+    Py_ssize_t nbytes, i;
+    char *data;
+
     if (!PyBytes_Check(bytes)) {
         PyErr_SetString(PyExc_TypeError, "bytes expected");
         return NULL;
     }
-    if (extend_bytes(self, bytes, BYTES_RAW) < 0)
+    nbytes = PyBytes_Size(bytes);
+    if (nbytes == 0)
+        Py_RETURN_NONE;
+
+    if (resize(self, self->nbits + nbytes) < 0)
         return NULL;
 
+    data = PyBytes_AsString(bytes);
+    for (i = 0; i < nbytes; i++)
+        setbit(self, self->nbits - nbytes + i, data[i] ? 1 : 0);
+
     Py_RETURN_NONE;
 }
 
@@ -1765,30 +1611,37 @@ transfer of data between bitarray objects to other python objects\n\
 static PyObject *
 bitarray_repr(bitarrayobject *self)
 {
-    PyObject *bytes;
-    PyObject *unpacked;
-#ifdef IS_PY3K
-    PyObject *decoded;
-#endif
+    PyObject *result;
+    Py_ssize_t i;
+    char *str;
+    size_t strsize;
 
-    if (self->nbits == 0) {
-        bytes = PyBytes_FromString("bitarray()");
+    if (self->nbits == 0)
+        return Py_BuildValue("s", "bitarray()");
+
+    strsize = self->nbits + 12;  /* 12 is the length of "bitarray('')" */
+    if (strsize > PY_SSIZE_T_MAX) {
+        PyErr_SetString(PyExc_OverflowError,
+                        "bitarray too large to represent");
+        return NULL;
     }
-    else {
-        bytes = PyBytes_FromString("bitarray(\'");
-        unpacked = unpack(self, '0', '1');
-        if (unpacked == NULL)
-            return NULL;
-        PyBytes_ConcatAndDel(&bytes, unpacked);
-        PyBytes_ConcatAndDel(&bytes, PyBytes_FromString("\')"));
+
+    str = (char *) PyMem_Malloc(strsize);
+    if (str == NULL) {
+        PyErr_NoMemory();
+        return NULL;
     }
-#ifdef IS_PY3K
-    decoded = PyUnicode_FromEncodedObject(bytes, NULL, NULL);
-    Py_DECREF(bytes);
-    return decoded;
-#else
-    return bytes;  /* really a string in Python 2 */
-#endif
+    /* add "bitarray('......')" to str */
+    strcpy(str, "bitarray('"); /* has length 10 */
+    str[strsize - 2] = '\'';
+    str[strsize - 1] = ')';
+
+    for (i = 0; i < self->nbits; i++)
+        str[i + 10] = GETBIT(self, i) ? '1' : '0';
+
+    result = Py_BuildValue("s#", str, (Py_ssize_t) strsize);
+    PyMem_Free((void *) str);
+    return result;
 }
 
 
@@ -2130,11 +1983,11 @@ bitarray_imul(bitarrayobject *self, PyObject *v)
 static PyObject *
 bitarray_cpinvert(bitarrayobject *self)
 {
-    PyObject *res;
+    PyObject *result;
 
-    res = bitarray_copy(self);
-    invert((bitarrayobject *) res);
-    return res;
+    result = bitarray_copy(self);
+    invert((bitarrayobject *) result);
+    return result;
 }
 
 #define BITWISE_FUNC(oper)  \
@@ -2822,7 +2675,8 @@ bitarray_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
                               endian_str == NULL ? np->endian : endian);
         if (a == NULL)
             return NULL;
-        memcpy(((bitarrayobject *) a)->ob_item, np->ob_item, Py_SIZE(np));
+        memcpy(((bitarrayobject *) a)->ob_item, np->ob_item,
+               (size_t) Py_SIZE(np));
 #undef np
         return a;
     }
@@ -2850,7 +2704,8 @@ bitarray_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
                                   endian);
             if (a == NULL)
                 return NULL;
-            memcpy(((bitarrayobject *) a)->ob_item, data + 1, nbytes - 1);
+            memcpy(((bitarrayobject *) a)->ob_item, data + 1,
+                   (size_t) nbytes - 1);
             return a;
         }
     }
@@ -2889,6 +2744,20 @@ richcompare(PyObject *v, PyObject *w, int op)
             Py_RETURN_TRUE;
     }
 
+    if (op == Py_EQ || op == Py_NE) {
+        assert(vs == ws);
+        /* shortcut for EQ/NE: if endianness is the same use memcmp()
+           We could use this shortcut for all sizes, but as we need to set
+           the unused bits, we shortcut only if we have at least one byte. */
+        if (vs >= 8 && va->endian == wa->endian) {
+            setunused(va);
+            setunused(wa);
+            cmp = memcmp(va->ob_item, wa->ob_item, (size_t) Py_SIZE(v));
+            return PyBool_FromLong((long) ((cmp ? 0 : 1) ^
+                                           ((op == Py_EQ) ? 0 : 1)));
+        }
+    }
+
     /* to avoid uninitialized warning for some compilers */
     vi = wi = 0;
     /* search for the first index where items are different */
@@ -3196,6 +3065,11 @@ bitdiff(PyObject *module, PyObject *args)
         return NULL;
     }
 
+    if (PyErr_WarnEx(PyExc_DeprecationWarning,
+                     "bitarray.bitdiff() has been deprecated since 1.2.0, "
+                     "use bitarray.util.count_xor() instead", 1) < 0)
+        return NULL;
+
 #define aa  ((bitarrayobject *) a)
 #define bb  ((bitarrayobject *) b)
     if (aa->nbits != bb->nbits) {
@@ -3305,13 +3179,7 @@ tuple(sizeof(void *),\n\
       sizeof(idx_t),\n\
       PY_SSIZE_T_MAX)");
 
-/*
-   In retrospect, I wish I had never added any modules functions (other
-   than get_default_endian() and _set_default_endian()) here.
-   These, and possibly many others should be part of a separate utility
-   module.  Anyway, at this point (2019) it is too late to remove them,
-   so I will just leave them here, but not any new ones.
-*/
+
 static PyMethodDef module_functions[] = {
     {"bitdiff",    (PyCFunction) bitdiff,    METH_VARARGS, bitdiff_doc   },
     {"bits2bytes", (PyCFunction) bits2bytes, METH_O,       bits2bytes_doc},
